// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    expr::{ops::VariantAccess, Expr, ToExpr},
    hdl,
    int::Bool,
    intern::{Intern, Interned},
    module::{
        enum_match_variants_helper, EnumMatchVariantAndInactiveScopeImpl,
        EnumMatchVariantsIterImpl, Scope,
    },
    source_location::SourceLocation,
    ty::{CanonicalType, MatchVariantAndInactiveScope, StaticType, Type, TypeProperties},
};
use hashbrown::HashMap;
use std::{fmt, iter::FusedIterator};

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct EnumVariant {
    pub name: Interned<str>,
    pub ty: Option<CanonicalType>,
}

impl EnumVariant {
    pub fn fmt_debug_in_enum(self) -> FmtDebugInEnum {
        FmtDebugInEnum(self)
    }
}

#[derive(Copy, Clone)]
pub struct FmtDebugInEnum(EnumVariant);

impl fmt::Debug for FmtDebugInEnum {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let EnumVariant { name, ty } = self.0;
        if let Some(ty) = ty {
            write!(f, "{name}({ty:?})")
        } else {
            write!(f, "{name}")
        }
    }
}

impl fmt::Display for FmtDebugInEnum {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        fmt::Debug::fmt(self, f)
    }
}

#[derive(Clone, Eq)]
struct EnumImpl {
    variants: Interned<[EnumVariant]>,
    name_indexes: HashMap<Interned<str>, usize>,
    type_properties: TypeProperties,
}

impl std::fmt::Debug for EnumImpl {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str("Enum ")?;
        f.debug_set()
            .entries(
                self.variants
                    .iter()
                    .map(|variant| variant.fmt_debug_in_enum()),
            )
            .finish()
    }
}

impl std::hash::Hash for EnumImpl {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.variants.hash(state);
    }
}

impl PartialEq for EnumImpl {
    fn eq(&self, other: &Self) -> bool {
        self.variants == other.variants
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub struct Enum(Interned<EnumImpl>);

impl fmt::Debug for Enum {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

const fn discriminant_bit_width_impl(variant_count: usize) -> usize {
    match variant_count.next_power_of_two().checked_ilog2() {
        Some(x) => x as usize,
        None => 0,
    }
}

#[derive(Clone)]
pub struct EnumTypePropertiesBuilder {
    type_properties: TypeProperties,
    variant_count: usize,
}

impl EnumTypePropertiesBuilder {
    #[must_use]
    pub const fn new() -> Self {
        Self {
            type_properties: TypeProperties {
                is_passive: true,
                is_storable: true,
                is_castable_from_bits: true,
                bit_width: 0,
            },
            variant_count: 0,
        }
    }
    pub const fn clone(&self) -> Self {
        Self { ..*self }
    }
    #[must_use]
    pub const fn variant(self, field_props: Option<TypeProperties>) -> Self {
        let Self {
            mut type_properties,
            variant_count,
        } = self;
        if let Some(TypeProperties {
            is_passive,
            is_storable,
            is_castable_from_bits,
            bit_width,
        }) = field_props
        {
            assert!(is_passive, "variant type must be a passive type");
            type_properties = TypeProperties {
                is_passive: true,
                is_storable: type_properties.is_storable & is_storable,
                is_castable_from_bits: type_properties.is_castable_from_bits
                    & is_castable_from_bits,
                bit_width: if type_properties.bit_width < bit_width {
                    bit_width
                } else {
                    type_properties.bit_width
                },
            };
        }
        Self {
            type_properties,
            variant_count: variant_count + 1,
        }
    }
    pub const fn finish(self) -> TypeProperties {
        assert!(
            self.variant_count != 0,
            "zero-variant enums aren't yet supported: \
            https://github.com/chipsalliance/firrtl-spec/issues/208",
        );
        let Some(bit_width) = self
            .type_properties
            .bit_width
            .checked_add(discriminant_bit_width_impl(self.variant_count))
        else {
            panic!("enum is too big: bit-width overflowed");
        };
        TypeProperties {
            bit_width,
            ..self.type_properties
        }
    }
}

impl Enum {
    #[track_caller]
    pub fn new(variants: Interned<[EnumVariant]>) -> Self {
        let mut name_indexes = HashMap::with_capacity(variants.len());
        let mut type_props_builder = EnumTypePropertiesBuilder::new();
        for (index, EnumVariant { name, ty }) in variants.iter().enumerate() {
            if let Some(old_index) = name_indexes.insert(*name, index) {
                panic!("duplicate variant name {name:?}: at both index {old_index} and {index}");
            }
            type_props_builder = type_props_builder.variant(ty.map(CanonicalType::type_properties));
        }
        Self(
            EnumImpl {
                variants,
                name_indexes,
                type_properties: type_props_builder.finish(),
            }
            .intern_sized(),
        )
    }
    pub fn discriminant_bit_width(self) -> usize {
        discriminant_bit_width_impl(self.variants().len())
    }
    pub fn type_properties(self) -> TypeProperties {
        self.0.type_properties
    }
    pub fn name_indexes(&self) -> &HashMap<Interned<str>, usize> {
        &self.0.name_indexes
    }
    pub fn can_connect(self, rhs: Self) -> bool {
        if self.0.variants.len() != rhs.0.variants.len() {
            return false;
        }
        for (
            &EnumVariant {
                name: lhs_name,
                ty: lhs_ty,
            },
            &EnumVariant {
                name: rhs_name,
                ty: rhs_ty,
            },
        ) in self.0.variants.iter().zip(rhs.0.variants.iter())
        {
            if lhs_name != rhs_name {
                return false;
            }
            match (lhs_ty, rhs_ty) {
                (None, None) => {}
                (None, Some(_)) | (Some(_), None) => return false,
                (Some(lhs_ty), Some(rhs_ty)) => {
                    if !lhs_ty.can_connect(rhs_ty) {
                        return false;
                    }
                }
            }
        }
        true
    }
}

pub trait EnumType:
    Type<
    BaseType = Enum,
    MaskType = Bool,
    MatchActiveScope = Scope,
    MatchVariantAndInactiveScope = EnumMatchVariantAndInactiveScope<Self>,
    MatchVariantsIter = EnumMatchVariantsIter<Self>,
>
{
    fn variants(&self) -> Interned<[EnumVariant]>;
    fn match_activate_scope(
        v: Self::MatchVariantAndInactiveScope,
    ) -> (Self::MatchVariant, Self::MatchActiveScope);
}

pub struct EnumMatchVariantAndInactiveScope<T: EnumType>(EnumMatchVariantAndInactiveScopeImpl<T>);

impl<T: EnumType> MatchVariantAndInactiveScope for EnumMatchVariantAndInactiveScope<T> {
    type MatchVariant = T::MatchVariant;
    type MatchActiveScope = Scope;

    fn match_activate_scope(self) -> (Self::MatchVariant, Self::MatchActiveScope) {
        T::match_activate_scope(self)
    }
}

impl<T: EnumType> EnumMatchVariantAndInactiveScope<T> {
    pub fn variant_access(&self) -> VariantAccess {
        self.0.variant_access()
    }
    pub fn activate(self) -> (VariantAccess, Scope) {
        self.0.activate()
    }
}

#[derive(Clone)]
pub struct EnumMatchVariantsIter<T: EnumType> {
    pub(crate) inner: EnumMatchVariantsIterImpl<T>,
    pub(crate) variant_index: std::ops::Range<usize>,
}

impl<T: EnumType> Iterator for EnumMatchVariantsIter<T> {
    type Item = EnumMatchVariantAndInactiveScope<T>;

    fn next(&mut self) -> Option<Self::Item> {
        self.variant_index.next().map(|variant_index| {
            EnumMatchVariantAndInactiveScope(self.inner.for_variant_index(variant_index))
        })
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        self.variant_index.size_hint()
    }
}

impl<T: EnumType> ExactSizeIterator for EnumMatchVariantsIter<T> {
    fn len(&self) -> usize {
        self.variant_index.len()
    }
}

impl<T: EnumType> FusedIterator for EnumMatchVariantsIter<T> {}

impl<T: EnumType> DoubleEndedIterator for EnumMatchVariantsIter<T> {
    fn next_back(&mut self) -> Option<Self::Item> {
        self.variant_index.next_back().map(|variant_index| {
            EnumMatchVariantAndInactiveScope(self.inner.for_variant_index(variant_index))
        })
    }
}

impl EnumType for Enum {
    fn match_activate_scope(
        v: Self::MatchVariantAndInactiveScope,
    ) -> (Self::MatchVariant, Self::MatchActiveScope) {
        let (expr, scope) = v.0.activate();
        (expr.variant_type().map(|_| expr.to_expr()), scope)
    }
    fn variants(&self) -> Interned<[EnumVariant]> {
        self.0.variants
    }
}

impl Type for Enum {
    type BaseType = Enum;
    type MaskType = Bool;
    type MatchVariant = Option<Expr<CanonicalType>>;
    type MatchActiveScope = Scope;
    type MatchVariantAndInactiveScope = EnumMatchVariantAndInactiveScope<Self>;
    type MatchVariantsIter = EnumMatchVariantsIter<Self>;

    fn match_variants(
        this: Expr<Self>,
        source_location: SourceLocation,
    ) -> Self::MatchVariantsIter {
        enum_match_variants_helper(this, source_location)
    }

    fn mask_type(&self) -> Self::MaskType {
        Bool
    }

    fn canonical(&self) -> CanonicalType {
        CanonicalType::Enum(*self)
    }

    #[track_caller]
    fn from_canonical(canonical_type: CanonicalType) -> Self {
        let CanonicalType::Enum(retval) = canonical_type else {
            panic!("expected enum");
        };
        retval
    }
    fn source_location() -> SourceLocation {
        SourceLocation::builtin()
    }
}

#[hdl]
pub enum HdlOption<T: Type> {
    HdlNone,
    HdlSome(T),
}

#[allow(non_snake_case)]
pub fn HdlNone<T: StaticType>() -> Expr<HdlOption<T>> {
    HdlOption[T::TYPE].HdlNone()
}

#[allow(non_snake_case)]
pub fn HdlSome<T: Type>(value: impl ToExpr<Type = T>) -> Expr<HdlOption<T>> {
    let value = value.to_expr();
    HdlOption[Expr::ty(value)].HdlSome(value)
}
